.. scikit-image
.. _`scikit-image`: https://scikit-image.org
.. _`scikit-image github`: https://github.com/scikit-image/scikit-image

.. _`scikit-image developer forum`: https://discuss.scientific-python.org/c/contributor/skimage
